/******************************************************************************
* x_idle_monitor.c - $Revision: 1.14 $ - $Date: 2004/06/19 20:27:43 $
*
*   Copyright (C) 2004 Craig B. Agricola (craig@theagricolas.org)
*
* This program started out just using the Xss (X Screensaver Extension) to
* track when I went idle and when I came back, printing everything out to
* the standard output.  This, of course, completely satisfied what I needed
* it for, but I couldn't leave well enough alone, so I just kept adding
* configurability and features to it.  Now it is a rather convenient program to
* start up in your login script or X startup script.   You'll probably even
* forget that it's there until either:
*
* 1) You are trying to figure out when you did something non-computer related,
*    and you realize that it corresponds directly to that span of idle time
*    in the log file
* 2) You run out of storage in your home directory because the
*    .x_idle_monitor.log file, which you had forgotten about, has grown over
*    the last few years to become enormous.  Obviously, a log-rotation setup
*    is probably the best bet to avoid this.  I'll avoid creeping-featuritis
*    by refusing to make that functionality part of this program...  I'll
*    consider it a *trivial* exercise for the reader to set this up.
*
* BUILDING
*
* This is how I build it with GCC under Linux with XFree86 libraries.  I'd love
* feedback on how it works with other compilers, operating systems, and X
* servers.
*
* gcc -O2 -o x_idle_monitor x_idle_monitor.c -L/usr/X11R6/lib -lX11 -lXext -lXss
*
* USAGE
*
* The --help option will give you a page describing the options and environment
* variables that you have access to, but in general, putting
* "x_idle_monitor --background"
* in your .xinitrc (or .xsession, or whatever you use to start the default
* programs under X) will set it up to log any idle periods of 5 minutes or
* more to $HOME/.x_idle_monitor.log.
*
* LICENSE
*
* x_idle_monitor is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation; either version 2, or (at your option) any later version.
*
* x_idle_monitor is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
* details.
*
* You should be able to download a copy of the GNU General Public License from
* http://www.gnu.org/licenses/gpl.txt
* If not, write to the Free Software Foundation,
* 59 Temple Place, Suite 330, Boston, MA 02111 USA.
*
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <pwd.h>
#include <X11/Xlib.h>
#include <X11/extensions/scrnsaver.h>

#include <getopt.h>

#define DEFAULT_IDLE_TIME_THRESH  300
#define NOW                         0

static char VersionString[] = "$Id: x_idle_monitor.c,v 1.14 2004/06/19 20:27:43 agricolc Exp $";
static char HelpString[] =
  "This program just uses the Xss (X Screen Saver) extension of the X\n"    \
  "server to poll for how long the user has been idle.  When the idle\n"    \
  "time reaches a configurable threshold, the last non-idle time is\n"      \
  "logged.  When the user becomes non-idle again, that time is logged\n"    \
  "as well.\n"                                                              \
  "\n"                                                                      \
  "Options:\n"                                                              \
  "    --logfile        (-l)    Specifies the log file to append to\n"      \
  "    --idle-threshold (-t)    Specifies the idle threshold in seconds\n"  \
  "    --background     (-b)    Runs in daemon mode in the background\n"    \
  "    --pidfile        (-p)    Lock file to use in daemon mode\n"          \
  "    --display        (-d)    Specifies the X server to connect to\n"     \
  "    --help           (-h)    Prints this help menu\n"                    \
  "    --version        (-v)    Prints the version string\n"                \
  "\n"                                                                      \
  "Environment Variables:\n"                                                \
  "    X_IDLE_MONITOR_LOG       Corresponds to the --logfile flag\n"        \
  "    X_IDLE_MONITOR_THRESH    Corresponds to the --idle-threshold flag\n" \
  "    X_IDLE_MONITOR_PIDFILE   Corresponds to the --pidfile flag\n"        \
  "    DISPLAY                  Corresponds to the --display flag\n"        \
  "\n"                                                                      \
  "Command line options always trump environment variable settings.\n"      \
  "\n"                                                                      \
  "In daemon mode, the PID file is first checked to see if there is an\n"   \
  "already running instance.  If the --pidfile option and\n"                \
  "X_IDLE_MONITOR_PIDFILE environment variable are not specified, the\n"    \
  "default file .x_idle_monitor.pid in the user's home directory will\n"    \
  "be used.  The point of the lock file for daemon mode is both to make\n"  \
  "it easy to find the daemon for termination, as well as to allow the\n"   \
  "user to start x_idle_monitor from a login script or startup script\n"    \
  "without having to worry if it gets called multiple times; only one\n"    \
  "instance will start.  If no log file is given in daemon mode, the\n"     \
  "default file .x_idle_monitor.log in the user's home directory will\n"    \
  "be used.";

static int   idle_time_thresh = DEFAULT_IDLE_TIME_THRESH;
static FILE *logfile = NULL;
static int   daemon_mode = 0;
static char *pidfile = NULL;
static char *display = NULL;

char *log_time(int adj) {
  time_t  t;
  char   *c, *i;
  time(&t);
  t -= adj;
  c = ctime(&t);
  for(i=c; (*i!='\n')&&(*i!=0); i++);
  if (*i=='\n') *i=0;
  return(c);
}

void log(int time_adjust, char *fmt, ...) {
  va_list arg_list;
  fprintf(logfile, "%s: ", log_time(time_adjust));
  va_start(arg_list, fmt);
  vfprintf(logfile, fmt, arg_list);
  va_end(arg_list);
  fprintf(logfile, "\n");
  fflush(logfile);
}

void die(char *fmt, ...) {
  va_list arg_list;
  va_start(arg_list, fmt);
  vfprintf(stderr, fmt, arg_list);
  va_end(arg_list);
  fprintf(stderr, "\n");
  exit(-1);
}

static int quit = 0;
void sig_hdl(int sig) {
  if (sig==SIGTSTP) {
    log(NOW, "Stopping idle monitoring");
    signal(SIGTSTP, SIG_DFL);
    kill(0, SIGTSTP);
  } else if (sig==SIGCONT) {
    signal(SIGTSTP, &sig_hdl);
    log(NOW, "Starting idle monitoring");
  } else if (sig==SIGPIPE) {
    log(NOW, "Stopping idle monitoring");
    if (pidfile) unlink(pidfile);
    exit(0);
  } else {
    quit=1;
  }
}
void trap_signals(void) {
  signal(SIGHUP, &sig_hdl);
  signal(SIGINT, &sig_hdl);
  signal(SIGQUIT, &sig_hdl);
  signal(SIGTERM, &sig_hdl);
  signal(SIGTSTP, &sig_hdl);
  signal(SIGCONT, &sig_hdl);
  signal(SIGPIPE, &sig_hdl);
}

char *get_default_file(char *ext) {
  static char *default_file_base = NULL;
  char *home;
  char *default_file;

  if (ext==NULL) return(NULL);
  if (default_file_base == NULL) {
    home = getenv("HOME");
    if (!home) {
      struct passwd *pw;
      pw = getpwuid(getuid());
      if (pw) home = pw->pw_dir;
    }
    if (home) {
      default_file_base = malloc(strlen(home) + 17);
      sprintf(default_file_base, "%s/.x_idle_monitor", home);
    } else
      default_file_base = "x_idle_monitor";
  }
  default_file = malloc(strlen(default_file_base)+strlen(ext)+2);
  sprintf(default_file, "%s.%s", default_file_base, ext);
  return(default_file);
}

void parse_options(int argc, char *argv[]) {
  char    opt;

  char   *logfile_str           = NULL;
  char   *logfile_from          = "logfile";
  char   *idle_time_thresh_str  = NULL;
  char   *idle_time_thresh_from = "idle-threshold";

  struct option long_opts[] = {
    { "logfile"       , required_argument, NULL, 'l' },
    { "idle-threshold", required_argument, NULL, 't' },
    { "background"    , no_argument      , NULL, 'b' },
    { "pidfile"       , required_argument, NULL, 'p' },
    { "display"       , required_argument, NULL, 'd' },
    { "help"          , no_argument      , NULL, 'h' },
    { "version"       , no_argument      , NULL, 'v' },
    { NULL, 0, NULL, 0 }
  };
  while ((opt=getopt_long(argc, argv, "l:t:bp:d:hv", long_opts, NULL)) != -1) {
    switch (opt) {
      case 'l':
        logfile_str = optarg;
        logfile_from = "--logfile option";
        break;
      case 't':
        idle_time_thresh_str = optarg;
        idle_time_thresh_from = "--idle-threshold option";
        break;
      case 'b':
        daemon_mode = 1;
        break;
      case 'p':
        pidfile = optarg;
        break;
      case 'd':
        display = optarg;
        break;
      case 'h':
        printf("\n%s\n\n", HelpString); exit(0); break;
      case 'v':
        printf("\nVersion: %s\n\n", VersionString); exit(0); break;
      default: exit(-1); break;
    }
  }

  if (!logfile_str) {
    logfile_str           = getenv("X_IDLE_MONITOR_LOG");
    logfile_from          = "X_IDLE_MONITOR_LOG environment variable";
  }
  if (!idle_time_thresh_str) {
    idle_time_thresh_str  = getenv("X_IDLE_MONITOR_THRESH");
    idle_time_thresh_from = "X_IDLE_MONITOR_THRESH environment variable";
  }
  if (!pidfile)   pidfile = getenv("X_IDLE_MONITOR_PIDFILE");
  if (!display)   display = getenv("DISPLAY");

  if ((!logfile_str) && (daemon_mode)) {
    logfile_str           = get_default_file("log");
    logfile_from          = "default";
  }

  if (logfile_str) {
    if (!(logfile = fopen(logfile_str, "a")))
      die("Couldn't open log file specifed by %s (%s): %s",
          logfile_from, logfile_str, strerror(errno));
  } else
    logfile = stdout;

  if (idle_time_thresh_str) {
    char *err;
    idle_time_thresh = strtol(idle_time_thresh_str, &err, 0);
    if (*err != 0)
      die("Couldn't parse %s value starting at \"%s\"",
          idle_time_thresh_from, err);
  } else
    idle_time_thresh = DEFAULT_IDLE_TIME_THRESH;

  if (!pidfile) pidfile = get_default_file("pid");
}

void handle_daemon_mode(void) {
  FILE  *pidfd;
  char   pid_str[32], *pp;
  pid_t  pid = 0;
  if (daemon_mode) {
    if (pidfd = fopen(pidfile, "r")) {
      fread(pid_str, 1, 31, pidfd);
      for(pp=pid_str; (*pp!=0)&&(*pp!='\n'); pp++);
      if (*pp=='\n') *pp=0;
      pid = strtol(pid_str, NULL, 0);
      fclose(pidfd);
    }
    if ((pid==0) || (kill(pid, 0) == -1)) {
      if (!(pidfd = fopen(pidfile, "w")))
        die("Couldn't open lock file to record PID...");
      if (fork()) { exit(0); }
      if (pidfd) {
        fprintf(pidfd, "%d\n", getpid());
        fclose(pidfd);
      }
      setpgrp();
      close(STDIN_FILENO);
      close(STDOUT_FILENO);
      close(STDERR_FILENO);
    } else { exit(0); }
  }
}

int main(int argc, char *argv[]) {
  Display *d;
  XScreenSaverInfo *i;
  int xss_event_basep, xss_error_basep;
  int gone_idle;

  parse_options(argc, argv);

  if (!(d = XOpenDisplay(display)))
    die("Couldn't open X display \"%s\"!", display);
  else if (!XScreenSaverQueryExtension(d, &xss_event_basep,
                                          &xss_error_basep))
    die("X Screen Saver extension is not available!");
  else if (!(i = XScreenSaverAllocInfo()))
    die("Failed to allocate X Screen Saver information memory!");
  else {
    handle_daemon_mode();
    trap_signals();
    log(NOW, "Starting idle monitoring");
    gone_idle = 0;
    while(!quit) {
      if (!XScreenSaverQueryInfo(d, DefaultRootWindow(d), i)) {
        fprintf(stderr, "Failed to get X Screen Saver information!\n");
        quit = 1;
      } else {
        int idle = i->idle / 1000;
        if ((gone_idle==0) && (idle >= idle_time_thresh)) {
          log(idle, "Gone idle");
          gone_idle = 1;
        } else if ((gone_idle==1) && (idle < idle_time_thresh)) {
          log(NOW, "Back from idle");
          gone_idle = 0;
        }
      }
      usleep(100);
    }
    log(NOW, "Stopping idle monitoring");
    if (pidfile) unlink(pidfile);
  }

  XCloseDisplay(d);
}
